#include "collide.h"

/*
 * Vectors
 */
cd_real m3_vector_square (cd_real x, cd_real y, cd_real z)
{
 return SQR(x) + SQR(y) + SQR(z);
}

void m3_normalize_vector (cd_real *x, cd_real *y, cd_real *z)
{
 double length = 1.0 / sqrt(m3_vector_square(*x, *y, *z));

 *x *= length;
 *y *= length;
 *z *= length;
}

cd_real m3_dot_product (cd_real x1, cd_real y1, cd_real z1,
                        cd_real x2, cd_real y2, cd_real z2)
{
 return ( (x1 * x2) + (y1 * y2) + (z1 * z2) );
}

void m3_cross_product (cd_real x1, cd_real y1, cd_real z1,
                       cd_real x2, cd_real y2, cd_real z2,
                       cd_real *xout, cd_real *yout, cd_real *zout)
{
 *xout = (y1 * z2) - (z1 * y2);
 *yout = (z1 * x2) - (x1 * z2);
 *zout = (x1 * y2) - (y1 * x2);
}

/*
 * Matrices
 */
void m3_transpose_rmat (m3_rmat *tm, m3_rmat *m)
{
 int i, j;

 if (tm == m) {
    cd_real t;
   
    for (i=0; i<3; i++) {
        for (j=i+1; j<3; j++) {
            t = (*m)[i][j];
            (*m)[i][j] = (*m)[j][i];
            (*m)[j][i] = t;
        }
    }
 } else {
    for (i=0; i<3; i++) {
        for (j=0; j<3; j++) {
            (*tm)[i][j] = (*m)[j][i];
        }
    }
 }
}

void m3_apply_rmat (m3_rmat *m, m3_vec v, m3_vec *w)
{
#ifndef OPENGL_MATRIX
 #define CALC_ROW(n)   (v.x * (*m)[n][0]) +\
                       (v.y * (*m)[n][1]) +\
                       (v.z * (*m)[n][2])
#else
 /* the OpenGL spec says the matrix be given in column major order */
 #define CALC_ROW(n)   (v.x * (*m)[0][n]) +\
                       (v.y * (*m)[1][n]) +\
                       (v.z * (*m)[2][n])
#endif

 w->x = CALC_ROW(0);
 w->y = CALC_ROW(1);
 w->z = CALC_ROW(2);

 #undef CALC_ROW
}

void m3_apply_xmat (m3_xmat *m, m3_vtx p, m3_vtx *q)
{
#ifndef OPENGL_MATRIX
 #define CALC_ROW(n)  ((p.x * m->r[n][0]) +\
                       (p.y * m->r[n][1]) +\
                       (p.z * m->r[n][2]) +\
                       m->t[n])
#else
 /* the OpenGL spec says the matrix be given in column major order */
 #define CALC_ROW(n)  ((p.x * m->r[0][n]) +\
                       (p.y * m->r[1][n]) +\
                       (p.z * m->r[2][n]) +\
                       m->t[n])
#endif

 q->x = CALC_ROW(0);
 q->y = CALC_ROW(1);
 q->z = CALC_ROW(2);

 #undef CALC_ROW
}

#define dcos(x) cos((x) * M_PI / 180.0)
#define dsin(x) sin((x) * M_PI / 180.0)

#define MAKE_ROTATION_f(x, y, z)           \
        cd_real sin_x = dsin(x);           \
        cd_real cos_x = dcos(x);           \
                                           \
        cd_real sin_y = dsin(y);           \
        cd_real cos_y = dcos(y);           \
                                           \
        cd_real sin_z = dsin(z);           \
        cd_real cos_z = dcos(z);           \
                                           \
        cd_real sinx_siny = sin_x * sin_y; \
        cd_real cosx_siny = cos_x * sin_y;

#define R00_f (cos_y * cos_z)
#define R10_f ((sinx_siny * cos_z) - (cos_x * sin_z))
#define R20_f ((cosx_siny * cos_z) + (sin_x * sin_z))

#define R01_f (cos_y * sin_z)
#define R11_f ((sinx_siny * sin_z) + (cos_x * cos_z))
#define R21_f ((cosx_siny * sin_z) - (sin_x * cos_z))

#define R02_f (-sin_y)
#define R12_f (sin_x * cos_y)
#define R22_f (cos_x * cos_y)

void m3_get_xmat (m3_xmat *m,
                  cd_real xrot, cd_real yrot, cd_real zrot,
                  cd_real x, cd_real y, cd_real z)
{
 MAKE_ROTATION_f(xrot, yrot, zrot);

 m->r[0][0] = R00_f;
 m->r[0][1] = R01_f;
 m->r[0][2] = R02_f;

 m->r[1][0] = R10_f;
 m->r[1][1] = R11_f;
 m->r[1][2] = R12_f;

 m->r[2][0] = R20_f;
 m->r[2][1] = R21_f;
 m->r[2][2] = R22_f;

 m->t[0] = x;
 m->t[1] = y;
 m->t[2] = z;
}
