/*
 * Collision detection:
 * Sample program for the SEPARATING VECTOR algorithm
 *
 *  Copyright (c) 2002 - Borca Daniel
 *  Email : dborca@yahoo.com
 *  Web   : http://www.geocities.com/dborca
 *
 *
 * This demo shows two polytopes on a DARKGRAY background. In case the two
 * polytopes collide, the background turns to DARKBLUE. If anything goes
 * wrong in the collision detection engine, the background becomes DARKRED.
 *
 * Special note: the collision detection transformation matrix must be the
 * same as the OpenGL transformation matrix (hence the Z, Y, X rotation).
 * Moreover, the transformation has to be applied in the same manner: column
 * major order; for this, OPENGL_MATRIX has to be defined in `collide.h'.
 */

#include <stdlib.h>

#include <GL/glut.h>

#define OPENGL_MATRIX

#include "coldet/collide.h"

m3_vtx cube[] = {
       {-3, -3, -3},
       {3, -3, -3},
       {3, -3, 3},
       {-3, -3, 3},
       {-3, 3, -3},
       {-3, 3, 3},
       {3, 3, 3},
       {3, 3, -3}
};
m3_vtx pyramid[] = {
       {0, 4, 0},
       {0, -4, -4},
       {-4, -4, 4},
       {4, -4, 4}
};

typedef struct {
        double xangle, yangle, zangle;
        double dist;
} locator;

locator camera = {
        0.0, 0.0, 0.0,
        20.0
};
locator cubepos = {
        45, 45, 45,
        2.5
};
locator pyramidpos = {
        -30, 30, 30,
        -2.5
};

const double angle_speed = 5.0;
const double dist_speed = 1.0;

static __inline void do_vertex (m3_vtx *v)
{
 glVertex3d(v->x, v->y, v->z);
}

static int collide (int Pn, m3_vtx *Pv, int Qn, m3_vtx *Qv, locator *Pl, locator *Ql)
{
 cd_bool r;
 m3_xmat Xp, Xq;
 m3_rmat invXp, invXq;

 m3_get_xmat(&Xp, Pl->xangle, Pl->yangle, Pl->zangle, Pl->dist, 0, 0);
 m3_get_xmat(&Xq, Ql->xangle, Ql->yangle, Ql->zangle, Ql->dist, 0, 0);
 m3_transpose_rmat(&invXp, &Xp.r);
 m3_transpose_rmat(&invXq, &Xq.r);

 r = separating_vector(Pn, Pv, Qn, Qv, &Xp, &Xq, &invXp, &invXq);

 if (cd_error==CD_OK) {
    return r;
 } else {
    return ~CD_TRUE;
 }
}

static void set_camera_position (void)
{
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 glFrustum(-1.0, 1.0, -1.0, 1.0, 1.0, 40.0);
 glTranslatef(0, 0, -camera.dist);
 glRotatef(camera.xangle, 1, 0, 0);
 glRotatef(camera.yangle, 0, 1, 0);
 glRotatef(camera.zangle, 0, 0, 1);
 glMatrixMode(GL_MODELVIEW);
}

static void display (void)
{
 int r = collide(sizeof(cube)/sizeof(m3_vtx), cube,
                 sizeof(pyramid)/sizeof(m3_vtx), pyramid,
                 &cubepos,
                 &pyramidpos);

 if (r==CD_FALSE) {
    glClearColor(0.25, 0.25, 0.25, 0);
 } else if (r==CD_TRUE) {
    glClearColor(0, 0, 0.5, 0);
 } else {
    glClearColor(0.5, 0, 0, 0);
 }

 set_camera_position();

 /* Clear the RGB buffer and the depth buffer */
 glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

 glLoadIdentity();
 /* Translate and rotate the object */
 glTranslatef(pyramidpos.dist, 0.0, 0.0);
 glRotatef(pyramidpos.zangle, 0.0, 0.0, 1.0);
 glRotatef(pyramidpos.yangle, 0.0, 1.0, 0.0);
 glRotatef(pyramidpos.xangle, 1.0, 0.0, 0.0);

 glColor3f(1.0, 0.0, 1.0);
 /* Draw the sides of the three-sided pyramid */
 glBegin(GL_TRIANGLE_FAN);
	do_vertex(&pyramid[0]);
	do_vertex(&pyramid[1]);
	do_vertex(&pyramid[2]);
	do_vertex(&pyramid[3]);
	do_vertex(&pyramid[1]);
 glEnd();

 glColor3f(0.0, 1.0, 1.0);
 /* Draw the base of the pyramid */
 glBegin(GL_TRIANGLES);
	do_vertex(&pyramid[1]);
	do_vertex(&pyramid[3]);
	do_vertex(&pyramid[2]);
 glEnd();

 glLoadIdentity();
 /* Translate for the cube */
 glTranslatef(cubepos.dist, 0.0, 0.0);
 glRotatef(cubepos.zangle, 0.0, 0.0, 1.0);
 glRotatef(cubepos.yangle, 0.0, 1.0, 0.0);
 glRotatef(cubepos.xangle, 1.0, 0.0, 0.0);

 glColor3f(0.0, 1.0, 0.0);
 /* Draw the sides of the cube */
 glBegin(GL_QUAD_STRIP);
	do_vertex(&cube[7]);
	do_vertex(&cube[1]);
	do_vertex(&cube[4]);
	do_vertex(&cube[0]);
	do_vertex(&cube[5]);
	do_vertex(&cube[3]);
	do_vertex(&cube[6]);
	do_vertex(&cube[2]);
	do_vertex(&cube[7]);
	do_vertex(&cube[1]);
 glEnd();

 glColor3f(0.0, 0.0, 1.0);
 /* Draw the top and bottom of the cube */
 glBegin(GL_QUADS);
	do_vertex(&cube[0]);
	do_vertex(&cube[1]);
	do_vertex(&cube[2]);
	do_vertex(&cube[3]);

	do_vertex(&cube[4]);
	do_vertex(&cube[5]);
	do_vertex(&cube[6]);
	do_vertex(&cube[7]);
 glEnd();

 glFlush();
 glutSwapBuffers();
}

static void keyboard (unsigned char key, int x, int y)
{
 switch (key) {
        case 27:
             exit(0);
        case 'w': case 'W':
             pyramidpos.xangle += angle_speed;
             break;
        case 's': case 'S':
             pyramidpos.xangle -= angle_speed;
             break;
        case 'a': case 'A':
             pyramidpos.yangle += angle_speed;
             break;
        case 'd': case 'D':
             pyramidpos.yangle -= angle_speed;
             break;
        case 'q': case 'Q':
             pyramidpos.dist -= dist_speed;
             break;
        case 'e': case 'E':
             pyramidpos.dist += dist_speed;
             break;
        default:
             return;
 }
 glutPostRedisplay();
}

static void special (int k, int x, int y)
{
 switch (k) {
        case GLUT_KEY_UP:
             camera.xangle += angle_speed;
             break;
        case GLUT_KEY_DOWN:
             camera.xangle -= angle_speed;
             break;
        case GLUT_KEY_LEFT:
             camera.yangle += angle_speed;
             break;
        case GLUT_KEY_RIGHT:
             camera.yangle -= angle_speed;
             break;
        case GLUT_KEY_PAGE_UP:
             camera.dist -= dist_speed;
             break;
        case GLUT_KEY_PAGE_DOWN:
             camera.dist += dist_speed;
             break;
        default:
             return;
 }
 glutPostRedisplay();
}

static void init (void)
{
 glShadeModel(GL_FLAT);
 glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
 glPolygonMode(GL_BACK, GL_POINTS);
 glEnable(GL_DEPTH_TEST);
 glCullFace(GL_BACK);
 glEnable(GL_CULL_FACE);
}

int main (int argc, char **argv)
{
 glutInit(&argc, argv);
 glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
 glutInitWindowSize (640, 480);
 glutInitWindowPosition (0, 0);
 glutCreateWindow (argv[0]);
 init ();
 glutDisplayFunc(display);
 glutKeyboardFunc(keyboard);
 glutSpecialFunc(special);
 glutMainLoop();
 return 0;
}
