/*
 * Collision detection:
 * Sample program for the SEPARATING VECTOR algorithm
 *
 *  Copyright (c) 2002 - Borca Daniel
 *  Email : dborca@yahoo.com
 *  Web   : http://www.geocities.com/dborca
 *
 *
 * Input filename syntax:
 *         Pn                                   // number of vertices
 *         x_1, y_1, z_1                        // vertex 1
 *         ...
 *         x_Pn                                 // vertex Pn
 *         Pxrot, Pyrot, Pzrot, Pxt, Pyt, Pzt   // rotation & translation
 *         Qn                                   // number of vertices
 *         x_1, y_1, z_1                        // vertex 1
 *         ...
 *         x_Qn                                 // vertex Qn
 *         Qxrot, Qyrot, Qzrot, Qxt, Qyt, Qzt   // rotation & translation
 *
 * Output destination: stdout
 */

#include <stdio.h>
#include <stdlib.h>

#include "coldet/collide.h"



int Pn, Qn;
m3_vtx *Pv, *Qv;
cd_real Pxrot, Pyrot, Pzrot, Pxt, Pyt, Pzt;
cd_real Qxrot, Qyrot, Qzrot, Qxt, Qyt, Qzt;



/*
 * Input
 */
static int read_input (char *filename)
{
 int i;
 FILE *f;

 if ((f=fopen(filename, "rt"))!=NULL) {
    if (fscanf(f, "%d\n", &Pn)) {
       if ((Pv=(m3_vtx *)malloc(Pn*sizeof(m3_vtx)))!=NULL) {
          for (i=0; i<Pn; i++) {
              if (fscanf(f, "%lf, %lf, %lf", &Pv[i].x, &Pv[i].y, &Pv[i].z)!=3) {
                 break;
              }
          }
          if ((i==Pn) && (fscanf(f, "%lf, %lf, %lf, %lf, %lf, %lf", &Pxrot, &Pyrot, &Pzrot, &Pxt, &Pyt, &Pzt)==6)) {
             if (fscanf(f, "%d\n", &Qn)) {
                if ((Qv=(m3_vtx *)malloc(Qn*sizeof(m3_vtx)))!=NULL) {
                   for (i=0; i<Qn; i++) {
                       if (fscanf(f, "%lf, %lf, %lf", &Qv[i].x, &Qv[i].y, &Qv[i].z)!=3) {
                          break;
                       }
                   }
                   if ((i==Qn) && (fscanf(f, "%lf, %lf, %lf, %lf, %lf, %lf", &Qxrot, &Qyrot, &Qzrot, &Qxt, &Qyt, &Qzt)==6)) {
                      fclose(f);
                      return 0;
                   }
                   free(Qv);
                }
             }
          }
          free(Pv);
       }
    }
    fclose(f);
 }

 return -1;
}



/*
 * Misc
 */
static void print_xmat (const char *s, m3_xmat *p)
{
 int i, j;

 puts(s);
 for (i=0; i<3; i++) {
     for (j=0; j<3; j++) {
         printf("%9.3f", p->r[i][j]);
         putchar((j<2) ? ' ' : '\n');
     }
 }
 for (j=0; j<3; j++) {
     printf("%9.3f", p->t[j]);
     putchar((j<2) ? ' ' : '\n');
 }
 putchar('\n');
}

static void print_polytope (const char *s, int n, m3_vtx *v, m3_xmat *m)
{
 int i;
 m3_vtx p;

 puts(s);
 for (i=0; i<n; i++) {
     m3_apply_xmat(m, v[i], &p);
     printf("%9.3f %9.3f %9.3f\n", p.x, p.y, p.z);
 }
 putchar('\n');
}



/*
 * Main
 */
int main (int argc, char **argv)
{
 cd_bool r;
 m3_xmat Xp, Xq;
 m3_rmat invXp, invXq;

 if (argc!=2) {
    printf("Usage: %s <filename>\n", argv[0]);
    return 0;
 }

 if (read_input(argv[1])==0) {
    m3_get_xmat(&Xp, Pxrot, Pyrot, Pzrot, Pxt, Pyt, Pzt);
    m3_get_xmat(&Xq, Qxrot, Qyrot, Qzrot, Qxt, Qyt, Qzt);
    m3_transpose_rmat(&invXp, &Xp.r);
    m3_transpose_rmat(&invXq, &Xq.r);

    print_xmat("P-matrix:", &Xp);
    print_polytope("P (transformed):", Pn, Pv, &Xp);
    print_xmat("Q-matrix:", &Xq);
    print_polytope("Q (transformed):", Qn, Qv, &Xq);

    r = separating_vector(Pn, Pv, Qn, Qv, &Xp, &Xq, &invXp, &invXq);

    printf("CD: ");
    switch (cd_error) {
           case CD_OK:
                if (r) {
                   printf("collision\n");
                } else {
                   printf("not collision\n");
                }
                break;
           case CD_CYCLE:
                printf("infinite cycling\n");
                break;
           case CD_HPERR:
                printf("internal failure\n");
                break;
           default:
                printf("undefined error\n");
    }
    return 0;
 } else {
    printf("Error 01: unable to read `%s'\n", argv[1]);
    return 1;
 }

 (void)argc;
 (void)argv;
}
